////////////////////////////////////////////////////////////////////////
/// \file WaveformProcessor.h
/// \brief Service for extracting timing and pulse height information
///        from traces with multipoint readout.
///
/// \author ram2aq@virginia.edu
////////////////////////////////////////////////////////////////////////
#ifndef NOVADDT_WAVEFORMPROCESSOR_H
#define NOVADDT_WAVEFORMPROCESSOR_H

#include "art/Framework/Principal/Run.h"
#include "art/Framework/Services/Registry/ActivityRegistry.h"
#include "art/Framework/Services/Registry/ServiceMacros.h"

#include "DAQDataFormats/RawMilliSlice.h"
#include "DAQDataFormats/RawNanoSlice.h"
#include "DDTBaseDataProducts/BaseProducts.h"
#include "NovaDAQConventions/DAQConventions.h"

#include <array>
#include <bitset>
#include <string>
#include <cmath>
#include <cassert>
#include "TFile.h"
#include "TTree.h"

namespace novaddt
{

    unsigned int nTableLoads = 0;

    // Bounds within which adc values must lie in order for them to be in our
    // table. adc1 must be less than adc3. adc2 is measured relative to
    // adc3. See genADCShapeFitLookupTable for more detail about the looping
    // order.

    const static int kADC3Max    = 512;
    const static int kADC3Min    = 256;
    const static int kADC1Min    = 180;
    const static int kADC1RelMax = 50;
    const static int kADC2Min    = 180;
    const static int kADC2Max    = 512;


    // To understand this you need to understand the constraints the looping in
    // genADCShapeFitLookupTable and be ready to think about the 3D shape this
    // produces in ADC space.

    const static unsigned int kNumTableEntries = 21545472;
      /* ((kADC3Min-kADC1Min)*(kADC3Min) + // Rectangular part, where adc1 < 0 */
      /*  ((kADC3Max-kADC3Min)*(kADC3Max-kADC3Min-1))/2)* // Triangular part, where 0 <= adc1 < adc0 */
      /* (kADC2Max-kADC2Min); // Thickness of the wedge due to adc2 */


  uint8_t ADCShapeFit(int16_t adc1, int16_t adc2, int16_t adc3,
                      double riseTime, double fallTime,
                      bool& goodTime);

  TDC OffsetToTDC(uint32_t const& t0, uint64_t const& high_time, 
                  uint8_t const& offset);

  inline double GetExpectations(double t0,
                                double riseTime, double fallTime,
                                const int16_t* obs,
                                double* exps);

  class WaveformProcessor
  {
  public:
     WaveformProcessor(fhicl::ParameterSet const& pset, 
                     art::ActivityRegistry& reg);
     ~WaveformProcessor();

     void postBeginRun(art::Run const& run);
     void reconfigure(const fhicl::ParameterSet& pset);

     void setup(daqdataformats::RawMicroSlice* theMicro, 
                daqdataformats::RawNanoSlice* theNano)
          { 
            fMicro = theMicro; fNano = theNano; 
            if (fUseMulti) initmulti();
            else initdcs(); 
          }

     void setup(uint32_t const& t0, uint64_t const& high_time, 
              int16_t const& adc1, int16_t const& adc2, int16_t const& adc3, 
              int16_t const& dcs);

     TDC getTDC() { return fTDC; }
     ADC getADC() { return fADC; }

  private:

     void process(uint32_t const& t0, uint64_t const& high_time, 
              int16_t const& adc1, int16_t const& adc2, int16_t const& adc3,
              ADC& adcpeak, bool& goodTime, TDC& retval);
     void initmulti();
     void initdcs();

     bool fTableLoaded = false;
     std::array<uint8_t, kNumTableEntries> fTable;

     //the adc part of the lookup table is not currently used,
     //but it is still part of the lookup table file
     std::array<uint16_t, kNumTableEntries> fADCTable;

     std::bitset<kNumTableEntries> fGood;
     std::string fShapeTableFilename;
     const double fSampleTime;
     const double fRiseTime;
     const double fFallTime;
     daqdataformats::RawMicroSlice* fMicro;
     daqdataformats::RawNanoSlice* fNano;
     const bool fUseMulti;
     const bool fUseMulti4ADC;
     const bool fDoHighADCFits;
     TDC fTDC;
     ADC fADC;
  };
}
DECLARE_ART_SERVICE(novaddt::WaveformProcessor, LEGACY)
#endif // NOVADDT_WAVEFORMPROCESSOR_H
////////////////////////////////////////////////////////////////////////
