////////////////////////////////////////////////////////////////////////
// Class:       ActivityTrigger
// Module Type: filter
// File:        ActivityTrigger_module.cc
//
// Generated at Wed Sep 12 16:15:01 2012 by Nova DDT using artmod
// from art v1_01_01.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Persistency/Common/Assns.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileService.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/TriggerDecision.h"
#include "DDTBaseDataProducts/GroupedHitList.h"
#include "DDTUtilities/AssociationUtil.h"

#include "DAQDataFormats/DAQDataFormats.h"

#include <string>
#include <algorithm>
#include <unordered_set>

// ROOT includes
#include <TTree.h>
#include <TH1I.h>

namespace novaddt {
  class ActivityTriggerAna;
}

class novaddt::ActivityTriggerAna : public art::EDFilter {

public:
  explicit ActivityTriggerAna(fhicl::ParameterSet const & p);
  virtual ~ActivityTriggerAna();

  virtual bool filter(art::Event & e);

  void beginJob() override;

private:

  std::string  _hitsModuleLabel;   ///< label of module making the HitList
  std::string  _hitsInstanceLabel; ///< instance label making the HitList
  unsigned int _minHits;           ///< Minimum number of hits in the slice
  unsigned int _minPlanesView;     ///< Minimum planes (unique) on each view                        
  unsigned int _minPlanes;         ///< Minimum planes (unique) the hits are spread in any view     
  unsigned int _maxNearWindow;     ///< Maximum window where to look for near planes                
  unsigned int _minNearPlanes;     ///< Minimum number of near planes                               
  unsigned int _prescale;          ///< holds prescaling variable from fcl file
  
  unsigned int _triggerCounts = 0;
  unsigned int _sliceCounts = 0;
  unsigned int _evtCounts = 0;

  art::ServiceHandle<art::TFileService> tfs;
  TTree *fTree;
  TH1I *fHisto;
  unsigned int fHits;
  unsigned int fXPlanes;
  unsigned int fYPlanes;
  unsigned int fPlanes;
  bool fContinuity;
};

//---------------------------------------------------------------------    
novaddt::ActivityTriggerAna::ActivityTriggerAna(fhicl::ParameterSet const & p)
  : _hitsModuleLabel  (p.get<std::string >("HitModuleLabel"  ))
  , _hitsInstanceLabel(p.get<std::string >("HitInstanceLabel"))
  , _minHits          (p.get<unsigned int>("MinHits"         ))
  , _minPlanesView    (p.get<unsigned int>("MinPlanesView"   ))
  , _minPlanes        (p.get<unsigned int>("MinPlanes"       ))
  , _maxNearWindow    (p.get<unsigned int>("MaxNearWindow"   ))
  , _minNearPlanes    (p.get<unsigned int>("MinNearPlanes"   ))
  , _prescale         (p.get<unsigned int>("prescale"        ))
{
  produces< std::vector<novaddt::TriggerDecision> >();
  //produces< art::Assns<novaddt::TriggerDecision, novaddt::HitList> >();
}

//---------------------------------------------------------------------    
novaddt::ActivityTriggerAna::~ActivityTriggerAna()
{
  // Clean up dynamic memory and other resources here.
}

void novaddt::ActivityTriggerAna::beginJob()
{
  // Output Histogram
  fHisto = tfs->make<TH1I>("fHisto","Histogram;ADCs;Entries",400,0,4000);
  fTree = tfs->make<TTree>("fTree","Output tree");
  fTree->Branch("Hits",&fHits);
  fTree->Branch("XPlanes",&fXPlanes);
  fTree->Branch("YPlanes",&fYPlanes);
  fTree->Branch("Planes",&fPlanes);
  fTree->Branch("Continuity",&fContinuity);
}

//---------------------------------------------------------------------    
bool novaddt::ActivityTriggerAna::filter(art::Event & e)
{
  //std::cout << "--- ActivityTriggerAna filter\n";
  ++_evtCounts;

  // Retrieve the (grouped) hit list for this event
  art::Handle<novaddt::GroupedHitList<novaddt::TDC> > hits;
  e.getByLabel(_hitsModuleLabel, _hitsInstanceLabel, hits);
  //std::cout << "\tgot "     << hits->n_groups()
  //          << " groups.\n";

  // Allocate the vector for the trigger decisions that we may make (and write to the event)
  std::unique_ptr<std::vector<novaddt::TriggerDecision>> td(new std::vector<novaddt::TriggerDecision>);

  std::unordered_set<int> XplaneList;
  std::unordered_set<int> YplaneList;
  std::unordered_set<int> planeList;

  // Loop over the slices in the event
  for(unsigned int i=0; i<hits->n_groups(); ++i){

    // Get the hits for the current slice
    novaddt::HitList theSliceHits = hits->get(i);

    XplaneList.clear();
    YplaneList.clear();
    planeList.clear();
    fContinuity = false;
    // Safety checks in case we mess up the fickle file
    if(_maxNearWindow < _minNearPlanes) fContinuity = true;
    if(!(_maxNearWindow * _minNearPlanes)) fContinuity = true;

    ++_sliceCounts;

    fHits = theSliceHits.size();
    for(unsigned int j=0; j < theSliceHits.size(); ++j){
      auto thePlane = theSliceHits.at(j).Plane().val;
      // Push the plane into the list of unique planes
      auto theView = theSliceHits.at(j).View().val;
      if(theView == daqchannelmap::X_VIEW) XplaneList.insert(thePlane);
      else if(theView == daqchannelmap::Y_VIEW) YplaneList.insert(thePlane);
      planeList.insert(thePlane);

      fHisto->Fill(theSliceHits.at(j).ADC().val);
    } // endfor j
      
    fXPlanes = XplaneList.size();
    fYPlanes = YplaneList.size();
    fPlanes = fXPlanes + fYPlanes;

    for(auto iPlane=planeList.begin(); iPlane!=planeList.end(); ++iPlane)
      {
	if(!fContinuity)
	  {
	    unsigned int fNearPlanes = 0;
	    for(unsigned int j=1; j<=_maxNearWindow; ++j)
	      fNearPlanes += planeList.count(*iPlane+j);

	    if(fNearPlanes >= _minNearPlanes)
	      fContinuity = true;
	  }
      } // end of continuity check 
 
    // Now the trigger requirements
    if(fHits >= _minHits)
      if(fPlanes >= _minPlanes)
	if(fXPlanes >= _minPlanesView)
	  if(fYPlanes >= _minPlanesView)
	    if(fContinuity)
	      ++_triggerCounts;

    fTree->Fill();
  } // end loop on hit groups

  bool goodTrigger = (td->size() > 0);

  e.put(std::move(td));

  std::cout << "Frequency: " << _triggerCounts*1./(_evtCounts*5.e-3) << " pm "
	    << sqrt(_triggerCounts*(_evtCounts-_triggerCounts)*1./_evtCounts)/(_evtCounts*5.e-3) << " Hz" << std::endl; // Binomial Error

  return goodTrigger;
}

DEFINE_ART_MODULE(novaddt::ActivityTriggerAna)
