#include "DDTBaseDataProducts/DAQHit.h"
#include "DAQUnpackUtils/DataCompression.h"

#include <algorithm>
#include <cstdio>

using namespace daqchannelmap;

novaddt::DAQHit::DAQHit() :
  _theDetID(novadaq::cnv::kUNKNOWN_DET),
  _theChannel(0),
  _thePlane(0),
  _theCell(0),
  _theView(0),
  _theADC(0),
  _theTDC(0),
  _FEBOverflowError(false)
{
}

novaddt::DAQHit::DAQHit(daqdataformats::RawMilliSlice* theMilli,
			daqdataformats::RawNanoSlice*  theNano) :
  _theDetID(theMilli->getHeader()->getDetId()),
  // encode the DAQ Format Channel (dchan)
  _theChannel(DAQChannelMap::getInstance(_theDetID)
	       ->encodeDChan(_theDetID,
			     theMilli->getHeader()->getDiblockId(),
			     theMilli->getHeader()->getDCMId(),
			     theNano->getHeader()->getFEBId(),
			     theNano->getHeader()->getPixelAddr())),
  // getPlane() requires Logical Format Channel (lchan)
  _thePlane(DAQChannelMap::getInstance(_theDetID)->getPlane
	    (DAQChannelMap::getInstance(_theDetID)->encodeLChan(_theChannel))),
  // getCell() requires lchan
  _theCell(DAQChannelMap::getInstance(_theDetID)->getCell
	   (DAQChannelMap::getInstance(_theDetID)->encodeLChan(_theChannel))),
  // getView() requires dchan
  _theView(DAQChannelMap::getInstance(_theDetID)->getView(_theChannel)),
  _theTDC(theNano->getTimeStamp()),
  _FEBOverflowError(theNano->getHeader()->getOverflowError()) 
{
  uint64_t time_high = theMilli->getFloatingMicroSlice()->getHighWord();
  
  // Combine the high 32 bits with the low 32 bits to 
  // form the 64-bit time:
  _theTDC = (time_high << 32) | theNano->getTimeStamp();
  
  _theADC = ADC_Extraction(theNano);
  
  if (_theADC.val > 4096)
    {
      std::cerr << "Invalid ADC Value: " << _theADC.val << std::endl;
      assert(false);
    }
}

novaddt::DAQHit::
DAQHit(novaddt::TDC tdc, novaddt::ADC adc, novaddt::Plane plane):
  _theDetID(novadaq::cnv::kUNKNOWN_DET),
  _theChannel(0),
  _thePlane(plane),
  _theCell(0),
  _theView(0),
  _theADC(adc),
  _theTDC(tdc),
  _FEBOverflowError(false)
{
}

novaddt::DAQHit::
DAQHit(int                  detID,
       daqchannelmap::dchan channel,
       novaddt::ADC         adc,
       novaddt::TDC         tdc):
  _theDetID(detID),
  // pass the dchan
  _theChannel(channel),
  // getPlane() requires lchan
  _thePlane(DAQChannelMap::getInstance(_theDetID)->getPlane
	    (DAQChannelMap::getInstance(_theDetID)->encodeLChan(_theChannel))),
  // getCell() requires lchan
  _theCell(DAQChannelMap::getInstance(_theDetID)->getCell
	   (DAQChannelMap::getInstance(_theDetID)->encodeLChan(_theChannel))),
  // getView() requires dchan
  _theView(DAQChannelMap::getInstance(_theDetID)->getView(_theChannel)),
  _theADC(adc),
  _theTDC(tdc),
  _FEBOverflowError(false)
{
}

novaddt::ADC
novaddt::DAQHit::ADC_Extraction(daqdataformats::RawNanoSlice* theNano) const
{
  /*
    This code is shamelessly copied and adapted from the offline 
    rawdata::RawDigit::ADC() method (see link below).  It interprets the 
    raw data format and extracts the ADC value.

    https://cdcvs.fnal.gov/redmine/projects/novaart/repository/entry/trunk/RawData/RawDigit.cxx#L66
  */

  const uint32_t version = theNano->getHeader()->getVersion();

  // non-multipoint nanoslice version
  if (version == 0)
    return theNano->getPulseHeight();


  static daqdataformats::NanoSliceVersionConvention conv;

  const unsigned int nSamples = conv.getNSamples(version);
  const unsigned int nPretrig = conv.getNPretriggeredSamples(version);

  // Otherwise it's raw and we need to calculate the DCS value for ourselves

  // Otherwise we don't have the peak value in our dataset
  assert(nSamples > nPretrig);

  // Look back 3 samples (1.5us in the FD)
  const unsigned int kDCSDistance = 3;

  // Otherwise we don't have the baseline point in our dataset
  assert(nPretrig >= kDCSDistance);

  // The DCS value is the peak value minus the baseline value
  return theNano->getValue(nPretrig) - 
    theNano->getValue(nPretrig - kDCSDistance);
}

novaddt::DCM novaddt::DAQHit::getDCM() const
{ 
  // getDCM() require dchan
  dcm_id_t dcm_id = 
    DAQChannelMap::getInstance(_theDetID)->getDCM(_theChannel);
  // getDiBlock require dchan
  int db_id = 
    DAQChannelMap::getInstance(_theDetID)->getDiBlock(_theChannel);

  DCM result(dcm_id + 12 * (db_id - 1));

  return result;
}

void novaddt::DAQHit::print() const
{
  // this method was undefined so I defined it to be a "do-nothing" 
  // method as it is referenced by HoughPoint/cxx/src/HoughPoint.cpp
  //
  // ...great programming...
  return;
}

void novaddt::DAQHit::reset()
{
  return;
}

/*
  ADC Conversion using Environment Switches

// decode the multipoint readout as if it were single point readout
#define USE_DCS_WAVEFORMS
#define DCS_TRIGGER_POINT 3

// unpack the entire wave form and then decode it using a lookup table 
// (available, but simplisticly implemented)
//#define USE_RAW_WAVEFORMS

// use this if we have to decompress the data (8 bit to 12 bit decomnpression)
//#define COMPRESSION_8BIT



novaddt::ADC
novaddt::DAQHit::ADC_Extraction(daqdataformats::RawNanoSlice* theNano) const
{
  if(version == 0){
    return theNano->getPulseHeight();
  }

  if(version > 0){

#ifdef USE_DCS_WAVEFORMS
    // This block unpacks the data into a DCS waveform
    int DCSform[20];                    // DCS baseline subtracted waveform
    const int DCSoffset=2;

    // The effective number of samples is the number of raw samples minus the DCS comparision offset
    const int effsamples = theNano->getNSamples() - DCSoffset;

    adc=0;
    for(int i=0; i< effsamples; ++i){
#ifdef COMPRESSION_8BIT
      DCSform[i]= DataCompression::decodeVal8Bit(theNano->getValue(i+DCSoffset)) 
	        - DataCompression::decodeVal8Bit(theNano->getValue(i));	    
#else
      DCSform[i]= theNano->getValue(i+DCSoffset) - theNano->getValue(i);
#endif
      // Return the maximum in the DCS profile
      if( adc < DCSform[i]){
	adc = DCSform[i];
      } //endif
    } // endfor i
#endif

#ifdef USE_RAW_WAVEFORMS
    int waveform[20];                   // Raw waveforms

    const int nsamples = theNano->getNSamples();
    for(int i=0; i< nsamples; ++i){
#ifdef COMPRESSION_8BIT
      waveform[i] = DataCompression::decodeVal8Bit(theNano->getValue(i));
#else
      waveform[i] = theNano->getValue(i);
#endif
    }
    // Now do a lookup table for the fit values
    // i.e. FitLookup(waveform, &pulseheight, &time)

    // For now we just pull out the Nth point
    adc = <waveform[DCS_TRIGGER_POINT];
#endif

  } // endif Version > 0

  return adc;
}

*/
