////////////////////////////////////////////////////////////////////////
// Class:       SendTriggerSignal
// Module Type: analyzer
// File:        SendTriggerSignal_module.cc
//
// Generated at Mon Nov 19 15:08:59 2012 by Martin Frank using artmod
// from art v1_01_01.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"

#include "DDTBaseDataProducts/GroupedHitList.h"
#include "DDTBaseDataProducts/TriggerDecision.h"

// directly from DDTSender:
#include <rms/provider/DDSConnection.h>
#include <rms/base/RmsDestination.h>
#include <rms/RmsSender.h>

#include <DAQDataFormats/RawTriggerHeader.h>
#include <DAQDataFormats/RawTriggerTime.h>
#include <DAQDataFormats/RawTriggerTimingMarker.h>
#include <DAQDataFormats/RawTriggerRange.h>
#include <DAQDataFormats/RawTriggerMask.h>
#include <DAQDataFormats/RawTrigger.h>

#include <NovaTimingUtilities/TimingUtilities.h>

#include <DAQDataFormats/TriggerDefines.h>
#include <DAQMessages/ccpp_GTMessages.h>
#include <DAQMessages/ccpp_DDTMessages.h>

// #include <NovaGlobalTrigger/GTConfig.h>
#include <NovaDAQUtilities/LocationUtils.h>

#include <rms/provider/DDSConnection.h>
#include <rms/base/RmsDestination.h>
#include <rms/RmsSender.h>

// Ganglia
#include <NovaDAQMonitorClient/NdmcClient.h>

using namespace gov::fnal::cd::rms;

namespace novaddt {
  class SendTriggerSignal;
}

class novaddt::SendTriggerSignal : public art::EDAnalyzer {
public:
  explicit SendTriggerSignal(fhicl::ParameterSet const & p);
  virtual ~SendTriggerSignal();

  virtual void analyze(art::Event const & e);
  // Ganglia master pointer.  Needs to be accessed from all over
  nova::novadaqmonitorclient::NdmcClient _daqMonClient; // daq monitor client
  void    startHeartbeat();              // start ganglia partition heartbeat
  void    stopHeartbeat();               // stop ganglia partition heartbeat

  double convertTDCtoSeconds(uint64_t tdc);
  
private:
  uint32_t cast_to_32_bits(uint64_t const& x) const;

  // for trigger:
  std::vector<std::string> trigger_labels_;

  // for messaging (RMS = Responsive Messaging System):
  bool send_trigger_messages_, verbose_;
  int rms_partition_;
  std::string rms_mailbox_;
  boost::shared_ptr<provider::DDSConnection> rms_connection_;
  base::RmsDestination rms_destination_;
  RmsSender<provider::DDSConnection, ddtmessages::DDTTriggerMessage> 
  rms_sender_;
  
  // Ganglia metrics
  bool make_ganglia_metrics_;
  int32_t OverallTriggerRate;
  std::map <std::string, int32_t> m_trigger_rates;  // a map of trigger rate metrics
  std::map <std::string, int32_t> m_trigger_lengths;// a map of trigger length metrics
  std::map <std::string, int32_t> m_trigger_delays; // a map of trigger delay metrics
  // Heartbeat for the monitor client
  int32_t _monIdPartitionHeartbeat;
};

novaddt::SendTriggerSignal::SendTriggerSignal(fhicl::ParameterSet const & p) :
  EDAnalyzer(p),
  trigger_labels_(p.get<std::vector<std::string>>("trigger_labels")),
  send_trigger_messages_(p.get<bool>("send_trigger_messages", true)),
  verbose_(p.get<bool>("verbose", false)),
  rms_partition_(p.get<int>("rms_partition_number")),
  rms_mailbox_(p.get<std::string>("rms_mailbox")),
  rms_connection_(new provider::DDSConnection
		  ("GlobalTriggerTestApplication", rms_partition_)),
  rms_destination_(rms_mailbox_,
		   base::RmsDestination::TRIGGER_CHANNEL),
  rms_sender_(rms_connection_, rms_destination_),
  make_ganglia_metrics_(p.get<bool>("make_ganglia_metrics", true))
{
  std::cout << "\nUsing Partition: " << rms_partition_ 
	    << "\nUsing Mailbox:   " << rms_mailbox_ << "\n" << std::endl;

  if (make_ganglia_metrics_){
      // set ganglia partition.
      _daqMonClient.setPartition(rms_partition_);
      _daqMonClient.setGroupName("DDT");
      // starting ganglia heartbeat
      startHeartbeat();
      
      OverallTriggerRate = _daqMonClient.createMetric<double>(
                            "Overall Trigger Rate", "Hz",
                            nova::novadaqmonitorclient::RATE, "",
                            "Combined rate of all triggers");
  } // end of make_ganglia_metrics
  
  std::cout << "\nSending Messages for the following trigger modules:" 
	    << std::endl;
  for (auto const & trig : trigger_labels_){
    std::cout << "\t" << trig << std::endl;
    if (make_ganglia_metrics_){
        // setup ganglia metrics
        // This would be nice to do using the names from TriggerDefines.h,
        // but at this stage all we have are the module name strings.
        // The modules themselves have trigger masks hard coded into them, so they're
        // impossible to access from here, only from the decisions they issue.
        // As a stop gap we can use the module label
        m_trigger_rates[trig] = _daqMonClient.createMetric<double>(
                                trig+" Trigger Rate", "Hz",
                                nova::novadaqmonitorclient::RATE, "",
                                "rate of the "+trig+" trigger");
                                
        m_trigger_lengths[trig] = _daqMonClient.createMetric<uint32_t>(
                                trig+" Trigger average length", "TDC",
                                nova::novadaqmonitorclient::AVERAGE, "",
                                "Average length of the "+trig+" trigger");
                                
        m_trigger_delays[trig] = _daqMonClient.createMetric<double>(
                                trig+" Trigger average delay", "ms",
                                nova::novadaqmonitorclient::AVERAGE, "",
                                "Average delay of the "+trig+" trigger");
    } // end of make_ganglia_metrics
  } // end of loop on trigger_labels_
}

novaddt::SendTriggerSignal::~SendTriggerSignal()
{
  try
  {
    rms_sender_.close();
    rms_connection_->close();
  } catch (...) {
    std::cerr << "\nAn exception occured when closeing the DDS connections.\n"
	      << std::endl;
  }
  if (make_ganglia_metrics_){
    // stop ganglia heartbeat
    stopHeartbeat();
  } // end of make_ganglia_metrics
}

void novaddt::SendTriggerSignal::analyze(art::Event const & e)
{
  
  for (auto const& trigger_label : trigger_labels_)
  {
    art::Handle<std::vector<TriggerDecision>> trigger_decisions;
    e.getByLabel(trigger_label, trigger_decisions);
    if (trigger_decisions.failedToGet()){
    if (verbose_) std::cout << "\tno trigger decision objects found for trigger module " << trigger_label << std::endl;
        continue;
    }
    if (verbose_ && trigger_decisions->size() > 0)
      std::cout << "\t\tsending " << trigger_decisions->size() 
		<< " trigger messages for trigger module " 
		<< trigger_label << std::endl;

    for (auto const& decision : *trigger_decisions)
    {
      ddtmessages::DDTTriggerMessage msg;

      // get the current time and convert it to NOvA time:
      // daqdataformats::RawTriggerTime triggerTime;
      timeval now;
      gettimeofday(&now,0);
      uint64_t nova_now;
      novadaq::timeutils::convertUnixTimeToNovaTime(now, nova_now);

      msg.trigger_type = decision.mask();
      msg.start_time = decision.start();
      msg.enabled = true;
      msg.decision_time = nova_now;

      // convert the 64 MHz TDC counts into 2 MHz (500 ns) time windows 
      // and round up by one:
      msg.duration = cast_to_32_bits(decision.duration() / 32 + 1);

      msg.time_to_decision = cast_to_32_bits(nova_now - decision.start());

      msg.prescale = decision.prescale();

      // the following values have to be obtained from the environment:
      msg.detector_id = 49;
      msg.buffer_node_id = daqdataformats::trigSourceID::TRIGSRC_ID_BUFFNODE_1;       msg.process_id = 51;

      if (send_trigger_messages_)
	rms_sender_.sendMessage(msg);
      if (make_ganglia_metrics_){
          // Send the trigger metrics to ganglia
          _daqMonClient.setValue(OverallTriggerRate,double(1.));
          _daqMonClient.setValue(m_trigger_rates[trigger_label],double(1.));
          _daqMonClient.setValue(m_trigger_lengths[trigger_label],(uint32_t)decision.duration());
          // convert the delay (which is in TDC) to ms
          double time_to_decisions_ms = convertTDCtoSeconds(nova_now - msg.start_time)*1000.;
          _daqMonClient.setValue(m_trigger_delays[trigger_label],time_to_decisions_ms);
      } // end of make_ganglia_metrics
    }
  }
}

uint32_t novaddt::SendTriggerSignal::cast_to_32_bits(uint64_t const& x) const
{
  uint32_t result;
  uint32_t down_caster = 0xFFFFFFFF;

  // Let's grab the lowe 32 bits of the 64-bit input:
  result = x & down_caster;

  return result;
}

void novaddt::SendTriggerSignal::startHeartbeat()
{
  std::string title ="";
  std::string description="Activity on partition (1=active/0=inactive).";
  _monIdPartitionHeartbeat=_daqMonClient.createMetric<double>("Activity","",nova::novadaqmonitorclient::HEARTBEAT,title,description);
}

void novaddt::SendTriggerSignal::stopHeartbeat()
{
  _daqMonClient.destroyMetric(_monIdPartitionHeartbeat); // silence heartbeat
  _monIdPartitionHeartbeat=-1;
}

double novaddt::SendTriggerSignal::convertTDCtoSeconds(uint64_t TDC)
{
    // TDC clock ticks at 64 MHz
    double one_tdc = 1. / 64e6;
    return double(TDC) * one_tdc; 
}

DEFINE_ART_MODULE(novaddt::SendTriggerSignal)
