////////////////////////////////////////////////////////////////////////
// Class:       Multiplet
// Module Type: filter
// File:        Multiplet_module.cc
//
// This is a slicer designed to keep only very small slices, with finding SN
// nu hits in mind.  It doesn't try to do 3D matching, just simply looks for
// things that are near in plane number and time.  Single hits are probably 
// best rejected: noise is predominantly singletons, even though so are almost
// half of your SN nu hits :(
//
// Generated at Tue Sep  9 04:40:50 2014 by Alec Habig using artmod
// from cetpkgsupport v1_07_00.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
//#include "art/Framework/Principal/Run.h"
//#include "art/Framework/Principal/SubRun.h"
//#include "art/Utilities/InputTag.h"
//#include "fhiclcpp/ParameterSet.h"
//#include "messagefacility/MessageLogger/MessageLogger.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/DAQHit.h"
//#include "DDTBaseDataProducts/CompareDAQHit.h"

#include "DAQChannelMap/DAQChannelMap.h"

#include <memory>

#include <list>

namespace novaddt {
  class Multiplet;
}

class novaddt::Multiplet : public art::EDFilter {
public:
  explicit Multiplet(fhicl::ParameterSet const & p);
  // The destructor generated by the compiler is fine for classes
  // without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  Multiplet(Multiplet const &) = delete;
  Multiplet(Multiplet &&) = delete;
  Multiplet & operator = (Multiplet const &) = delete;
  Multiplet & operator = (Multiplet &&) = delete;

  // Required functions.
  bool filter(art::Event & event) override;

  // Selected optional functions.
  void beginJob() override;
  void endJob() override;

private:

  // Declare member data here.
  std::string     fInputLabel;
  std::string     fInputInstance;
  int             fMaxPlaneSep;
  int             fMaxTDCSep;
  uint            fMinSliceSize;
  uint            fMaxSliceSize;
  bool            fVerbose;
  void findslices(art::Handle<novaddt::HitList>  & inhits, std::unique_ptr< std::vector< novaddt::HitList > >& product);
  // Counters
  int _nEvents        = 0;
  int _nHitsIn        = 0;
  int _nHitsOut       = 0;
  int _nSlicesOut       = 0;
};


novaddt::Multiplet::Multiplet(fhicl::ParameterSet const & p):
fInputLabel     (p.get< std::string       >("InputLabel"   )),
fInputInstance  (p.get< std::string       >("InputInstance")),
fMaxPlaneSep    (p.get< int               >("MaxPlaneSep"  )),
fMaxTDCSep      (p.get< int               >("MaxTDCSep"    )),
fMinSliceSize   (p.get< int               >("MinSliceSize" )),
fMaxSliceSize   (p.get< int               >("MaxSliceSize" )),
fVerbose        (p.get< bool              >("Verbose"      ))
// Initialize member data
{
    std::cout << "--- novaddt::Multiplet begin" << std::endl;
    std::cout << "\t Input hitlists label:    " << fInputLabel    << std::endl;
    std::cout << "\t Input hitlists instance: " << fInputInstance << std::endl;
    std::cout << "\t Max Z separation:        " << fMaxPlaneSep   << std::endl;
    std::cout << "\t Max time separation:     " << fMaxTDCSep     << std::endl;
    std::cout << "\t Min slice size:          " << fMinSliceSize  << std::endl;
    std::cout << "\t Max slice size:          " << fMaxSliceSize  << std::endl;
    std::cout << "\t Verbose:                 " << fVerbose       << std::endl;
    
    // Call appropriate Produces<>() functions
    produces<std::vector<HitList>>("Multiplets");
}
//////////////////////////////////////////////////////////////////////

bool novaddt::Multiplet::filter(art::Event & event)
{
    _nEvents++;
    
    // Hitlist from event (input)
    art::Handle<novaddt::HitList> hits;
    event.getByLabel(fInputLabel, fInputInstance, hits);

    _nHitsIn+=hits->size();

    //make a vector of hitlists out of hits after slicing
    std::unique_ptr< std::vector<HitList> >product(new std::vector<HitList>);

    // Find slices
    findslices(hits,product);
    
    // Add filtered hitlists to event
    _nSlicesOut+=product->size();
    for (auto slice: *product) _nHitsOut+=slice.size();
    event.put(std::move(product),"Multiplets");

    return true;
}

void novaddt::Multiplet::beginJob()
{
  // Implementation of optional member function here.
}
//////////////////////////////////////////////////////////////////////
void novaddt::Multiplet::endJob()
{
    std::cout << "--- novaddt::Multiplet end " << std::endl;
    std::cout << "\t# of events:        " << _nEvents   << std::endl;
    std::cout << "\t# of hits input:    " << _nHitsIn   << std::endl;
    std::cout << "\t# of hits output:   " << _nHitsOut  << std::endl;
    std::cout << "\t# of slices output: " << _nSlicesOut  << std::endl;
}

//////////////////////////////////////////////////////////////////////

void novaddt::Multiplet::findslices(art::Handle<novaddt::HitList> & inhits, std::unique_ptr<std::vector<HitList> > & product)
{
  if (!inhits->empty()) {

    // copy input hitlist to a list that we can start slicing up
    std::list<DAQHit> hits(inhits->begin(),inhits->end());
    if (fVerbose) std::cout << "findslices got " << inhits->size() << " hits, working with " << hits.size()<<std::endl;

    // Loop over all hits
    while (!hits.empty()) {
      // this will be the first hit in the slice, put it there
      HitList frag;
      frag.emplace_back(hits.front());
      // remove it from the front of the list, since it's now in a slice
      hits.pop_front();
      // point to the first hit in the slice
      auto lastInSlice = frag.back();
      // point to next hit in the list
      auto testhit = hits.begin();
      // Find hits associated in space and time with hit testpoint
      while (testhit != hits.end()) {
	// since hits are time ordered, march along until we get out of time,
	// with the last hit already in the slice, adding hits that are 
	// also in space with the last hit already in the slice
	int deltaTDC = abs((*testhit).TDC().val - lastInSlice.TDC().val);
	if (deltaTDC <= fMaxTDCSep) { 
	  // hit in time, see if in space too
	  int deltaPlane = abs((*testhit).Plane().val - lastInSlice.Plane().val);
	  //	  std::cout << "TDC: " << (*testhit).TDC().val << "(" << deltaTDC 
	  //		    << ") Plane: " << (*testhit).Plane().val 
	  //		    << "(" << deltaPlane << ")"<< std::endl;
	  if (deltaPlane <= fMaxPlaneSep) {
	    // add this hit to slice
	    frag.emplace_back(*testhit);
	    lastInSlice = frag.back();
	    // remove it from list of potential hits, since a hit can only 
	    // belong to one slice (after moving to next hit)
	    auto erasehit = testhit;
	    testhit++;
	    hits.erase(erasehit);
	  } else {// if deltaPlane ok
	    // go look at next hit anyway
	    testhit++;
	  }
	} else break; // if deltaTDC ok look more close, else done with slice
      } // end loop over rest of hits

      // we've finished building a slice, save it to product
      if (fVerbose) std::cout << "Sliced: " << frag.size() << " hits from " 
		<< frag.front().TDC().val << " to "
		<< frag.back().TDC().val <<std::endl;
      if ((frag.size() <= fMaxSliceSize) 
	  && (frag.size() >= fMinSliceSize)) {
	product->emplace_back(frag);
      }
    } // loop over full hitlist

  }
  // Last slices seem to be noise slices.  
  // Save the slices that don't meet the cuts in here
  // TODO: right now we're just dropping "noise" hits, more efficient.  
  // Make an empty slice just in case
  HitList noise;
  product->emplace_back(noise);
}
//////////////////////////////////////////////////////////////////////
DEFINE_ART_MODULE(novaddt::Multiplet)
